open OUnit
open Constraint

let raises_exception ?(exn) f =
  match exn with
  |Some e ->
    (try f (); false
    with e' when e=e' -> true)
  |None   ->
    (try f (); false
    with _ -> true)

module Values =
  struct
    open Indexes
    include Indexes_test.Values
    let int1_le_varn    = bounds_constraint int1 Le termvarn
    let int1_gt_varn    = bounds_constraint int1 Gt termvarn
    let int1_le_nplus1  = bounds_constraint int1 Le nplus1
    let int1_le_nminus1 = bounds_constraint int1 Le nminus1
    let nplus1_le_int1  = bounds_constraint nplus1 Le int1
    let nplus1_gt_int1  = bounds_constraint nplus1 Gt int1
    let n'plus1_gt_int1 = bounds_constraint n'plus1 Gt int1
    let int2_le_n       = bounds_constraint int2 Le termvarn
    let bot             = bounds_conjunction int1_le_varn int1_gt_varn
    let top             = Constraint.top
    let n_eq_1          = constraint_atom termvarn Eq int1
  end

open Values

let test =
  "constraints" >:::
    ["equal" >::
      (fun () ->
        "1<=n 1<=n"    @? (equal int1_le_varn int1_le_varn);
        "1<=n 1>n"     @? (not (equal int1_le_varn int1_gt_varn));
        "1>n 1<=n"     @? (not (equal int1_gt_varn int1_le_varn));
        "1>n 1>n"      @? (equal int1_gt_varn int1_gt_varn);
        "n+1>1 n+1>1"  @? (equal nplus1_gt_int1 nplus1_gt_int1);
        "n+1>1 n+1<=1" @? (not (equal nplus1_gt_int1 nplus1_le_int1));
        "top top"      @? (equal top top);
        "bot bot"      @? (equal bot bot);
        "top bot"      @? (not (equal top bot));
        "bot top"      @? (not (equal bot top));
      );
      "bounds_constraint" >::
      (fun () ->
        "1<=2"     @? raises_exception ~exn:NonRegularBounds (fun () -> bounds_constraint int1 Le int2);
        "1>2"      @? raises_exception ~exn:NonRegularBounds (fun () -> bounds_constraint int1 Gt int2);
        "n>n'"     @? raises_exception ~exn:NonRegularBounds (fun () -> bounds_constraint termvarn Gt termvarn');
        "n+1>n'+2" @? raises_exception ~exn:NonRegularBounds (fun () -> bounds_constraint nplus1 Gt n'plus2);
        "n<=n"     @? raises_exception ~exn:NonRegularBounds (fun () -> bounds_constraint termvarn Le termvarn);
        "n>n"      @? raises_exception ~exn:NonRegularBounds (fun () -> bounds_constraint termvarn Gt termvarn);
        "n+1<=n"   @? raises_exception ~exn:NonRegularBounds (fun () -> bounds_constraint nplus1 Le termvarn);
        "n+1>n"    @? raises_exception ~exn:NonRegularBounds (fun () -> bounds_constraint nplus1 Gt termvarn);
        "n+1<=n+2" @? raises_exception ~exn:NonRegularBounds (fun () -> bounds_constraint nplus1 Le nplus2);
        "n+1>n+2"  @? raises_exception ~exn:NonRegularBounds (fun () -> bounds_constraint nplus1 Gt nplus2);
        "n<=n+1"   @? raises_exception ~exn:NonRegularBounds (fun () -> bounds_constraint nplus1 Le nplus1);
        "n>n+3"    @? raises_exception ~exn:NonRegularBounds (fun () -> bounds_constraint nplus1 Gt (nplus3));
      );
    "bounds_conjunction" >::
      (fun () ->
        "top&top"     @? (bounds_conjunction top top = top);
        "int1_le_varn&top"   @? (bounds_conjunction int1_le_varn top = int1_le_varn);
        "int1_gt_varn&top"   @? (bounds_conjunction int1_gt_varn top = int1_gt_varn);
        "int1_le_nplus1&top"   @? (bounds_conjunction int1_le_nplus1 top = int1_le_nplus1);
        "int1_le_nminus1&top"   @? (bounds_conjunction int1_le_nminus1 top = int1_le_nminus1);
        "nplus1_le_int1&top"   @? (bounds_conjunction nplus1_le_int1 top = nplus1_le_int1);
        "nplus1_gt_int1&top"   @? (bounds_conjunction nplus1_gt_int1 top = nplus1_gt_int1);
        "top&int1_le_varn"   @? (bounds_conjunction top int1_le_varn = int1_le_varn);
        "top&int1_gt_varn"   @? (bounds_conjunction top int1_gt_varn = int1_gt_varn);
        "top&int1_le_nplus1"   @? (bounds_conjunction top int1_le_nplus1 = int1_le_nplus1);
        "top&int1_le_nminus1"   @? (bounds_conjunction top int1_le_nminus1 = int1_le_nminus1);
        "top&nplus1_le_int1"   @? (bounds_conjunction top nplus1_le_int1 = nplus1_le_int1);
        "top&nplus1_gt_int1"   @? (bounds_conjunction top nplus1_gt_int1 = nplus1_gt_int1);
        "int1_le_varn&int1_gt_varn" @? (is_false (bounds_conjunction int1_le_varn int1_gt_varn));
        "int1_le_varn&int1_gt_varn" @? (is_false (bounds_conjunction int1_gt_varn int1_le_varn));
        "int1_le_varn&int1_le_nplus1" @? (bounds_conjunction int1_le_varn int1_le_nplus1 = int1_le_varn);
        "int1_le_nplus1&int1_le_varn" @? (bounds_conjunction int1_le_nplus1 int1_le_varn = int1_le_varn);
        "int1_le_varn&int1_le_nminus1" @? (bounds_conjunction int1_le_varn int1_le_nminus1 = int1_le_nminus1);
        "int1_le_nminus1&int1_le_varn" @? (bounds_conjunction int1_le_nminus1 int1_le_varn = int1_le_nminus1);
        "int1_le_nplus1&int1_le_nminus1" @? (bounds_conjunction int1_le_nplus1 int1_le_nminus1 = int1_le_nminus1);
        "int1_le_nminus1&int1_le_nplus1" @? (bounds_conjunction int1_le_nminus1 int1_le_nplus1 = int1_le_nminus1);
        "int1_le_varn&n'plus1_gt_int1" @? raises_exception (fun () -> bounds_conjunction int1_le_varn n'plus1_gt_int1);
        "int1_le_varn&n'plus1_gt_int1" @? raises_exception (fun () -> bounds_conjunction n'plus1_gt_int1 int1_le_varn);
      );
    "always_greater" >::
      (fun () ->
        "top->"        @? (not (always_greater ~hyp:top int1 termvarn));
        "top->"        @? (not (always_greater ~hyp:top int1 nplus1));
        "top->"        @? (not (always_greater ~hyp:top int1 nminus1));
        "top->"        @? (not (always_greater ~hyp:top nplus1 int1));
        "bot->"        @? (always_greater ~hyp:bot int1 termvarn);
        "bot->"        @? (always_greater ~hyp:bot int1 nplus1);
        "bot->"        @? (always_greater ~hyp:bot int1 nminus1);
        "bot->"        @? (always_greater ~hyp:bot nplus1 int1);
        "n>0->n>0"     @? (always_greater ~hyp:int1_le_varn termvarn int0);
        "n>0->n>-1"    @? (always_greater ~hyp:int1_le_varn termvarn minus1);
        "n>1->n>0"     @? (always_greater ~hyp:int2_le_n termvarn int0);
        "not n>0->n>1" @? (not (always_greater ~hyp:int1_le_varn termvarn int1));
        "1>n->1>n"     @? (always_greater ~hyp:int1_gt_varn int1 termvarn);
        "1>n->2>n"     @? (always_greater ~hyp:int1_gt_varn int2 termvarn);
        "1>n->1>n-1"   @? (always_greater ~hyp:int1_gt_varn int2 nminus1);
        "not 1>n->0>n" @? (not (always_greater ~hyp:int1_gt_varn int0 termvarn));
      );
    (*
     *"always_equal" >::
     *  (fun () ->
     *    "bot->"        @? (always_equal ~hyp:bot int1 termvarn);
     *    "bot->"        @? (always_equal ~hyp:bot int1 termvarn);
     *    "bot->"        @? (always_equal ~hyp:bot int1 nplus1);
     *    "bot->"        @? (always_equal ~hyp:bot int1 nminus1);
     *    "bot->"        @? (always_equal ~hyp:bot nplus1 int1);
     *    "top->"        @? (not (always_equal ~hyp:top int1 termvarn));
     *    "top->"        @? (not (always_equal ~hyp:top int1 nplus1));
     *    "top->"        @? (not (always_equal ~hyp:top int1 nminus1));
     *    "top->"        @? (not (always_equal ~hyp:top nplus1 int1));
     *    "n=n"          @? (always_equal ~hyp:top termvarn termvarn);
     *    "1=1"          @? (always_equal ~hyp:top int1 int1);
     *    "n+1=n+1"      @? (always_equal ~hyp:top nplus1 nplus1);
     *    "n=1->n=1"     @? (always_equal ~hyp:n_eq_1 termvarn int1);
     *    "n>0/->n=1"    @? (not (always_equal ~hyp:int1_le_varn termvarn int1));
     *  );
     *)
    "always_different" >::
      (fun () ->
        "bot->"      @? (always_different ~hyp:bot int1 termvarn);
        "bot->"      @? (always_different ~hyp:bot int1 nplus1);
        "bot->"      @? (always_different ~hyp:bot int1 nminus1);
        "bot->"      @? (always_different ~hyp:bot nplus1 int1);
        "1=1"        @? (not (always_different ~hyp:top int1 int1));
        "n=n"        @? (not (always_different ~hyp:top termvarn termvarn));
        "t=t"        @? (not (always_different ~hyp:top nplus1 nplus1));
        "!n=1"       @? (not (always_different ~hyp:top int1 termvarn));
        "!nplus1=1"  @? (not (always_different ~hyp:top int1 nplus1));
        "!t=1"       @? (not (always_different ~hyp:top int1 nminus1));
        "!1=nplus1"  @? (not (always_different ~hyp:top nplus1 int1));
        "1<>0"       @? (always_different ~hyp:top int1 int0);
        "n+1<>n"     @? (always_different ~hyp:top nplus1 termvarn);
        "n<>n+1"     @? (always_different ~hyp:top termvarn nplus1);
        "n+2<>n+1"   @? (always_different ~hyp:top nplus2 nplus1);
        "n=1->n<>2"  @? (always_different ~hyp:n_eq_1 termvarn int2);
        "n=1->n<>0"  @? (always_different ~hyp:n_eq_1 termvarn int0);
        "n>1->n<>1"  @? (always_different ~hyp:int2_le_n termvarn int1);
        "n>1->n<>0"  @? (always_different ~hyp:int2_le_n termvarn int0);
        "n>1->!n<>2" @? (not (always_different ~hyp:int2_le_n termvarn int2));
        "n>0->n<>0"  @? (always_different ~hyp:int1_le_varn termvarn int0);
      );
    ]
