open Core.Std open OUnit open Literal

let raises_exception f =
  try f (); false
  with _ -> true

module Values = 
  struct
    include Indexes_test.Values
    let p  = Proposition.of_string "P"
    let q  = Proposition.of_string "Q"
    let p1 = pos p int1
    let p2 = pos p int2
    let q1 = pos q int1
    let n_p1 = neg p int1
    let pn = pos p termvarn
    let pnplus1 = pos p nplus1
    let pnminus1 = pos p nminus1
    let n_pnplus1 = neg p nplus1
  end

open Values

let test =
  "literals" >:::
    [ "equal" >::
      (fun () ->
        "p1 p1"   @? (equal p1 p1);
        "pn pn"   @? (equal pn pn);
        "~p1 ~p1" @? (equal n_p1 n_p1);
        "p1 pn"   @? (not (equal p1 pn));
        "pn p1"   @? (not (equal pn p1));
      );
      "looping map" >:::
      let add' m ~key = LoopingMap.add m ~key ~data:() in
      let empty_set = LoopingMap.empty in
      let p1_set = add'        empty_set ~key:p1 in
      let p2_set = add'        empty_set ~key:p2 in
      let q1_set = add'        empty_set ~key:q1 in
      let n_p1_set = add'      empty_set ~key:n_p1 in
      let pn_set = add'        empty_set ~key:pn in
      let pnplus1_set = add'   empty_set ~key:pnplus1 in
      let pnminus1_set = add'  empty_set ~key:pnminus1 in
      let n_pnplus1_set = add' empty_set ~key:n_pnplus1 in
      let set1 = add' p1_set ~key:p2 in
      let set2 = add' set1 ~key:pn in
      let set3 = add' set2 ~key:pnminus1 in
      [
        "singletons" >::
          (fun () ->
            "p1  -> p1"               @? (Option.is_some (LoopingMap.find p1_set ~key:p1));
            "p2  -> p2"               @? (Option.is_some (LoopingMap.find p2_set ~key:p2));
            "q1  -> q1"               @? (Option.is_some (LoopingMap.find q1_set ~key:q1));
            "n_p1  -> n_p1"           @? (Option.is_some (LoopingMap.find n_p1_set ~key:n_p1));
            "pn  -> pn"               @? (Option.is_some (LoopingMap.find pn_set ~key:pn));
            "pnplus1  -> pnplus1"     @? (Option.is_some (LoopingMap.find pnplus1_set ~key:pnplus1));
            "pnminus1  -> pnminus1"   @? (Option.is_some (LoopingMap.find pnminus1_set ~key:pnminus1));
            "n_pnplus1  -> n_pnplus1" @? (Option.is_some (LoopingMap.find n_pnplus1_set ~key:n_pnplus1));
          );
        "sets" >::
          (fun () ->
            "p1   -> {p1;p2}"         @? (Option.is_some (LoopingMap.find set1 ~key:p1));
            "p2   -> {p1;p2}"         @? (Option.is_some (LoopingMap.find set1 ~key:p2));
            "p1   -> {p1;p2;pn}"      @? (Option.is_some (LoopingMap.find set2 ~key:p1));
            "p2   -> {p1;p2;pn}"      @? (Option.is_some (LoopingMap.find set2 ~key:p2));
            "pn   -> {p1;p2;pn}"      @? (Option.is_some (LoopingMap.find set2 ~key:pn));
            "p1   -> {p1;p2;pn;pn-1}" @? (Option.is_some (LoopingMap.find set3 ~key:p1));
            "p2   -> {p1;p2;pn;pn-1}" @? (Option.is_some (LoopingMap.find set3 ~key:p2));
            "pn   -> {p1;p2;pn;pn-1}" @? (Option.is_some (LoopingMap.find set3 ~key:pn));
            "pn-1 -> {p1;p2;pn;pn-1}" @? (Option.is_some (LoopingMap.find set3 ~key:pnminus1));
            "pn+1 -> {p1;p2;pn;pn-1}" @? (not (Option.is_some (LoopingMap.find set3 ~key:pnplus1)));
            "q1   -> {p1;p2;pn;pn-1}" @? (not (Option.is_some (LoopingMap.find set3 ~key:q1)));
          );
     ]
    ]
